@@ -13,7 +13,7 @@ use Data::Object::ClassHas;
 
 extends 'Zing::Store';
 
-our $VERSION = '0.02'; # VERSION
+our $VERSION = '0.03'; # VERSION
 
 # ATTRIBUTES
 
@@ -44,6 +44,16 @@ fun new_client($self) {
   );
 }
 
+has meta => (
+  is => 'ro',
+  isa => 'Str',
+  new => 1,
+);
+
+fun new_meta($self) {
+  require Zing::ID; Zing::ID->new->string
+}
+
 has table => (
   is => 'ro',
   isa => 'Str',
@@ -63,13 +73,14 @@ fun new_encoder($self) {
 fun BUILD($self) {
   my $client = $self->client;
   my $table = $self->table;
-  do {
+  local $@; eval {
     $client->do(qq{
       create table if not exists "$table" (
         "id" serial primary key,
         "key" varchar not null,
         "value" text not null,
-        "index" integer default 0
+        "index" integer default 0,
+        "meta" varchar null
       )
     });
   }
@@ -93,6 +104,8 @@ fun DESTROY($self) {
 
 # METHODS
 
+my $retries = 10;
+
 method drop(Str $key) {
   my $table = $self->table;
   my $client = $self->client;
@@ -118,16 +131,33 @@ method keys(Str $query) {
 method lpull(Str $key) {
   my $table = $self->table;
   my $client = $self->client;
+  for my $attempt (1..$retries) {
+    local $@; eval {
+      my $sth = $client->prepare(
+        qq{
+          update "$table" set "meta" = ? where "id" = (
+            select "me"."id" from "$table" "me"
+            where "me"."key" = ? and "me"."meta" is null
+            order by "me"."index" asc limit 1
+          )
+        }
+      );
+      $sth->execute($self->meta, $key);
+    };
+    if ($@) {
+      die $@ if $attempt == $retries;
+    }
+    else {
+      last;
+    }
+  }
   my $data = $client->selectrow_arrayref(
     qq{
-      select "t0"."id", "t0"."value"
-      from "$table" "t0" where "t0"."id" = (
-        select min("t1"."id")
-        from "$table" "t1" where "t1"."key" = ?
-      )
+      select "id", "value"
+      from "$table" where "meta" = ? and "key" = ? order by "index" asc limit 1
     },
     {},
-    $key,
+    $self->meta, $key,
   );
   if ($data) {
     my $sth = $client->prepare(
@@ -149,7 +179,17 @@ method lpush(Str $key, HashRef $val) {
       ))
     }
   );
-  $sth->execute($key, $self->encode($val), $key);
+  for my $attempt (1..$retries) {
+    local $@; eval {
+      $sth->execute($key, $self->encode($val), $key);
+    };
+    if ($@) {
+      die $@ if $attempt == $retries;
+    }
+    else {
+      last;
+    }
+  }
   return $sth->rows;
 }
 
@@ -175,16 +215,33 @@ method recv(Str $key) {
 method rpull(Str $key) {
   my $table = $self->table;
   my $client = $self->client;
+  for my $attempt (1..$retries) {
+    local $@; eval {
+      my $sth = $client->prepare(
+        qq{
+          update "$table" set "meta" = ? where "id" = (
+            select "me"."id" from "$table" "me"
+            where "me"."key" = ? and "me"."meta" is null
+            order by "me"."index" desc limit 1
+          )
+        }
+      );
+      $sth->execute($self->meta, $key);
+    };
+    if ($@) {
+      die $@ if $attempt == $retries;
+    }
+    else {
+      last;
+    }
+  }
   my $data = $client->selectrow_arrayref(
     qq{
-      select "t0"."id", "t0"."value"
-      from "$table" "t0" where "t0"."id" = (
-        select max("t1"."id")
-        from "$table" "t1" where "t1"."key" = ?
-      )
+      select "id", "value"
+      from "$table" where "meta" = ? and "key" = ? order by "index" desc limit 1
     },
     {},
-    $key,
+    $self->meta, $key,
   );
   if ($data) {
     my $sth = $client->prepare(
@@ -206,7 +263,17 @@ method rpush(Str $key, HashRef $val) {
       ))
     }
   );
-  $sth->execute($key, $self->encode($val), $key);
+  for my $attempt (1..$retries) {
+    local $@; eval {
+      $sth->execute($key, $self->encode($val), $key);
+    };
+    if ($@) {
+      die $@ if $attempt == $retries;
+    }
+    else {
+      last;
+    }
+  }
   return $sth->rows;
 }
 
